@@ -1,19 +1,36 @@
 import sys
 from collections import defaultdict
+from copy import deepcopy
 from enum import Enum
-from typing import Any, Callable, ClassVar, DefaultDict, Dict, List, Mapping, Optional, Type, get_type_hints
+from typing import (
+    Any,
+    Callable,
+    ClassVar,
+    Counter,
+    DefaultDict,
+    Dict,
+    List,
+    Mapping,
+    Optional,
+    Type,
+    TypeVar,
+    get_type_hints,
+)
 from uuid import UUID, uuid4
 
 import pytest
 
 from pydantic import (
+    BaseConfig,
     BaseModel,
     ConfigError,
     Extra,
     Field,
     NoneBytes,
     NoneStr,
+    PrivateAttr,
     Required,
+    SecretStr,
     ValidationError,
     constr,
     root_validator,
@@ -191,6 +208,16 @@ def test_allow_extra():
     assert Model(a='10.2', b=12).dict() == {'a': 10.2, 'b': 12}
 
 
+def test_allow_extra_repr():
+    class Model(BaseModel):
+        a: float = ...
+
+        class Config:
+            extra = Extra.allow
+
+    assert str(Model(a='10.2', b=12)) == 'a=10.2 b=12'
+
+
 def test_forbidden_extra_success():
     class ForbiddenExtra(BaseModel):
         foo = 'whatever'
@@ -283,9 +310,15 @@ def test_set_attr_invalid():
 def test_any():
     class AnyModel(BaseModel):
         a: Any = 10
+        b: object = 20
 
-    assert AnyModel().a == 10
-    assert AnyModel(a='foobar').a == 'foobar'
+    m = AnyModel()
+    assert m.a == 10
+    assert m.b == 20
+
+    m = AnyModel(a='foobar', b='barfoo')
+    assert m.a == 'foobar'
+    assert m.b == 'barfoo'
 
 
 def test_alias():
@@ -1348,7 +1381,72 @@ def test_model_iteration():
     assert dict(m) == {'c': 3, 'd': Foo()}
 
 
-def test_model_export_nested_list():
+@pytest.mark.parametrize(
+    'exclude,expected,raises_match',
+    [
+        pytest.param(
+            {'foos': {0: {'a'}, 1: {'a'}}},
+            {'c': 3, 'foos': [{'b': 2}, {'b': 4}]},
+            None,
+            id='excluding fields of indexed list items',
+        ),
+        pytest.param(
+            {'foos': {'a'}},
+            TypeError,
+            'expected integer keys',
+            id='should fail trying to exclude string keys on list field (1).',
+        ),
+        pytest.param(
+            {'foos': {0: ..., 'a': ...}},
+            TypeError,
+            'expected integer keys',
+            id='should fail trying to exclude string keys on list field (2).',
+        ),
+        pytest.param(
+            {'foos': {0: 1}},
+            TypeError,
+            'Unexpected type',
+            id='should fail using integer key to specify list item field name (1)',
+        ),
+        pytest.param(
+            {'foos': {'__all__': 1}},
+            TypeError,
+            'Unexpected type',
+            id='should fail using integer key to specify list item field name (2)',
+        ),
+        pytest.param(
+            {'foos': {'__all__': {'a'}}},
+            {'c': 3, 'foos': [{'b': 2}, {'b': 4}]},
+            None,
+            id='using "__all__" to exclude specific nested field',
+        ),
+        pytest.param(
+            {'foos': {0: {'b'}, '__all__': {'a'}}},
+            {'c': 3, 'foos': [{}, {'b': 4}]},
+            None,
+            id='using "__all__" to exclude specific nested field in combination with more specific exclude',
+        ),
+        pytest.param(
+            {'foos': {'__all__'}},
+            {'c': 3, 'foos': []},
+            None,
+            id='using "__all__" to exclude all list items',
+        ),
+        pytest.param(
+            {'foos': {1, '__all__'}},
+            {'c': 3, 'foos': []},
+            None,
+            id='using "__all__" and other items should get merged together, still excluding all list items',
+        ),
+        pytest.param(
+            {'foos': {1: {'a'}, -1: {'b'}}},
+            {'c': 3, 'foos': [{'a': 1, 'b': 2}, {}]},
+            None,
+            id='using negative and positive indexes, referencing the same items should merge excludes',
+        ),
+    ],
+)
+def test_model_export_nested_list(exclude, expected, raises_match):
     class Foo(BaseModel):
         a: int = 1
         b: int = 2
@@ -1359,41 +1457,273 @@ def test_model_export_nested_list():
 
     m = Bar(c=3, foos=[Foo(a=1, b=2), Foo(a=3, b=4)])
 
-    assert m.dict(exclude={'foos': {0: {'a'}, 1: {'a'}}}) == {'c': 3, 'foos': [{'b': 2}, {'b': 4}]}
-
-    with pytest.raises(TypeError, match='expected integer keys'):
-        m.dict(exclude={'foos': {'a'}})
-    with pytest.raises(TypeError, match='expected integer keys'):
-        m.dict(exclude={'foos': {0: ..., 'a': ...}})
-    with pytest.raises(TypeError, match='Unexpected type'):
-        m.dict(exclude={'foos': {0: 1}})
-    with pytest.raises(TypeError, match='Unexpected type'):
-        m.dict(exclude={'foos': {'__all__': 1}})
-
-    assert m.dict(exclude={'foos': {0: {'b'}, '__all__': {'a'}}}) == {'c': 3, 'foos': [{}, {'b': 4}]}
-    assert m.dict(exclude={'foos': {'__all__': {'a'}}}) == {'c': 3, 'foos': [{'b': 2}, {'b': 4}]}
-    assert m.dict(exclude={'foos': {'__all__'}}) == {'c': 3, 'foos': []}
-
-    with pytest.raises(ValueError, match='set with keyword "__all__" must not contain other elements'):
-        m.dict(exclude={'foos': {1, '__all__'}})
+    if isinstance(expected, type) and issubclass(expected, Exception):
+        with pytest.raises(expected, match=raises_match):
+            m.dict(exclude=exclude)
+    else:
+        original_exclude = deepcopy(exclude)
+        assert m.dict(exclude=exclude) == expected
+        assert exclude == original_exclude
 
 
-def test_model_export_dict_exclusion():
+@pytest.mark.parametrize(
+    'excludes,expected',
+    [
+        pytest.param(
+            {'bars': {0}},
+            {'a': 1, 'bars': [{'y': 2}, {'w': -1, 'z': 3}]},
+            id='excluding first item from list field using index',
+        ),
+        pytest.param({'bars': {'__all__'}}, {'a': 1, 'bars': []}, id='using "__all__" to exclude all list items'),
+        pytest.param(
+            {'bars': {'__all__': {'w'}}},
+            {'a': 1, 'bars': [{'x': 1}, {'y': 2}, {'z': 3}]},
+            id='exclude single dict key from all list items',
+        ),
+    ],
+)
+def test_model_export_dict_exclusion(excludes, expected):
     class Foo(BaseModel):
         a: int = 1
         bars: List[Dict[str, int]]
 
     m = Foo(a=1, bars=[{'w': 0, 'x': 1}, {'y': 2}, {'w': -1, 'z': 3}])
 
-    excludes = {'bars': {0}}
-    assert m.dict(exclude=excludes) == {'a': 1, 'bars': [{'y': 2}, {'w': -1, 'z': 3}]}
-    assert excludes == {'bars': {0}}
-    excludes = {'bars': {'__all__'}}
-    assert m.dict(exclude=excludes) == {'a': 1, 'bars': []}
-    assert excludes == {'bars': {'__all__'}}
-    excludes = {'bars': {'__all__': {'w'}}}
-    assert m.dict(exclude=excludes) == {'a': 1, 'bars': [{'x': 1}, {'y': 2}, {'z': 3}]}
-    assert excludes == {'bars': {'__all__': {'w'}}}
+    original_excludes = deepcopy(excludes)
+    assert m.dict(exclude=excludes) == expected
+    assert excludes == original_excludes
+
+
+def test_model_exclude_config_field_merging():
+    """Test merging field exclude values from config."""
+
+    class Model(BaseModel):
+        b: int = Field(2, exclude=...)
+
+        class Config:
+            fields = {
+                'b': {'exclude': ...},
+            }
+
+    assert Model.__fields__['b'].field_info.exclude is ...
+
+    class Model(BaseModel):
+        b: int = Field(2, exclude={'a': {'test'}})
+
+        class Config:
+            fields = {
+                'b': {'exclude': ...},
+            }
+
+    assert Model.__fields__['b'].field_info.exclude == {'a': {'test'}}
+
+    class Model(BaseModel):
+        b: int = Field(2, exclude={'foo'})
+
+        class Config:
+            fields = {
+                'b': {'exclude': {'bar'}},
+            }
+
+    assert Model.__fields__['b'].field_info.exclude == {'foo': ..., 'bar': ...}
+
+
+def test_model_exclude_copy_on_model_validation():
+    """When `Config.copy_on_model_validation` is set, it should keep private attributes and excluded fields"""
+
+    class User(BaseModel):
+        _priv: int = PrivateAttr()
+        id: int
+        username: str
+        password: SecretStr = Field(exclude=True)
+        hobbies: List[str]
+
+    my_user = User(id=42, username='JohnDoe', password='hashedpassword', hobbies=['scuba diving'])
+
+    my_user._priv = 13
+    assert my_user.id == 42
+    assert my_user.password.get_secret_value() == 'hashedpassword'
+    assert my_user.dict() == {'id': 42, 'username': 'JohnDoe', 'hobbies': ['scuba diving']}
+
+    class Transaction(BaseModel):
+        id: str
+        user: User = Field(..., exclude={'username'})
+        value: int
+
+        class Config:
+            fields = {'value': {'exclude': True}}
+
+    t = Transaction(
+        id='1234567890',
+        user=my_user,
+        value=9876543210,
+    )
+
+    assert t.user is not my_user
+    assert t.user.hobbies == ['scuba diving']
+    assert t.user.hobbies is my_user.hobbies  # `Config.copy_on_model_validation` only does a shallow copy
+    assert t.user._priv == 13
+    assert t.user.password.get_secret_value() == 'hashedpassword'
+    assert t.dict() == {'id': '1234567890', 'user': {'id': 42, 'hobbies': ['scuba diving']}}
+
+
+@pytest.mark.parametrize(
+    'kinds',
+    [
+        {'sub_fields', 'model_fields', 'model_config', 'sub_config', 'combined_config'},
+        {'sub_fields', 'model_fields', 'combined_config'},
+        {'sub_fields', 'model_fields'},
+        {'combined_config'},
+        {'model_config', 'sub_config'},
+        {'model_config', 'sub_fields'},
+        {'model_fields', 'sub_config'},
+    ],
+)
+@pytest.mark.parametrize(
+    'exclude,expected',
+    [
+        (None, {'a': 0, 'c': {'a': [3, 5], 'c': 'foobar'}, 'd': {'c': 'foobar'}}),
+        ({'c', 'd'}, {'a': 0}),
+        ({'a': ..., 'c': ..., 'd': {'a': ..., 'c': ...}}, {'d': {}}),
+    ],
+)
+def test_model_export_exclusion_with_fields_and_config(kinds, exclude, expected):
+    """Test that exporting models with fields using the export parameter works."""
+
+    class ChildConfig:
+        pass
+
+    if 'sub_config' in kinds:
+        ChildConfig.fields = {'b': {'exclude': ...}, 'a': {'exclude': {1}}}
+
+    class ParentConfig:
+        pass
+
+    if 'combined_config' in kinds:
+        ParentConfig.fields = {
+            'b': {'exclude': ...},
+            'c': {'exclude': {'b': ..., 'a': {1}}},
+            'd': {'exclude': {'a': ..., 'b': ...}},
+        }
+
+    elif 'model_config' in kinds:
+        ParentConfig.fields = {'b': {'exclude': ...}, 'd': {'exclude': {'a'}}}
+
+    class Sub(BaseModel):
+        a: List[int] = Field([3, 4, 5], exclude={1} if 'sub_fields' in kinds else None)
+        b: int = Field(4, exclude=... if 'sub_fields' in kinds else None)
+        c: str = 'foobar'
+
+        Config = ChildConfig
+
+    class Model(BaseModel):
+        a: int = 0
+        b: int = Field(2, exclude=... if 'model_fields' in kinds else None)
+        c: Sub = Sub()
+        d: Sub = Field(Sub(), exclude={'a'} if 'model_fields' in kinds else None)
+
+        Config = ParentConfig
+
+    m = Model()
+    assert m.dict(exclude=exclude) == expected, 'Unexpected model export result'
+
+
+def test_model_export_exclusion_inheritance():
+    class Sub(BaseModel):
+        s1: str = 'v1'
+        s2: str = 'v2'
+        s3: str = 'v3'
+        s4: str = Field('v4', exclude=...)
+
+    class Parent(BaseModel):
+        a: int
+        b: int = Field(..., exclude=...)
+        c: int
+        d: int
+        s: Sub = Sub()
+
+        class Config:
+            fields = {'a': {'exclude': ...}, 's': {'exclude': {'s1'}}}
+
+    class Child(Parent):
+        class Config:
+            fields = {'c': {'exclude': ...}, 's': {'exclude': {'s2'}}}
+
+    actual = Child(a=0, b=1, c=2, d=3).dict()
+    expected = {'d': 3, 's': {'s3': 'v3'}}
+    assert actual == expected, 'Unexpected model export result'
+
+
+def test_model_export_with_true_instead_of_ellipsis():
+    class Sub(BaseModel):
+        s1: int = 1
+
+    class Model(BaseModel):
+        a: int = 2
+        b: int = Field(3, exclude=True)
+        c: int = Field(4)
+        s: Sub = Sub()
+
+        class Config:
+            fields = {'c': {'exclude': True}}
+
+    m = Model()
+    assert m.dict(exclude={'s': True}) == {'a': 2}
+
+
+def test_model_export_inclusion():
+    class Sub(BaseModel):
+        s1: str = 'v1'
+        s2: str = 'v2'
+        s3: str = 'v3'
+        s4: str = 'v4'
+
+    class Model(BaseModel):
+        a: Sub = Sub()
+        b: Sub = Field(Sub(), include={'s1'})
+        c: Sub = Field(Sub(), include={'s1', 's2'})
+
+        class Config:
+            fields = {'a': {'include': {'s2', 's1', 's3'}}, 'b': {'include': {'s1', 's2', 's3', 's4'}}}
+
+    Model.__fields__['a'].field_info.include == {'s1': ..., 's2': ..., 's3': ...}
+    Model.__fields__['b'].field_info.include == {'s1': ...}
+    Model.__fields__['c'].field_info.include == {'s1': ..., 's2': ...}
+
+    actual = Model().dict(include={'a': {'s3', 's4'}, 'b': ..., 'c': ...})
+    # s1 included via field, s2 via config and s3 via .dict call:
+    expected = {'a': {'s3': 'v3'}, 'b': {'s1': 'v1'}, 'c': {'s1': 'v1', 's2': 'v2'}}
+
+    assert actual == expected, 'Unexpected model export result'
+
+
+def test_model_export_inclusion_inheritance():
+    class Sub(BaseModel):
+        s1: str = Field('v1', include=...)
+        s2: str = Field('v2', include=...)
+        s3: str = Field('v3', include=...)
+        s4: str = 'v4'
+
+    class Parent(BaseModel):
+        a: int
+        b: int
+        c: int
+        s: Sub = Field(Sub(), include={'s1', 's2'})  # overrides includes set in Sub model
+
+        class Config:
+            # b will be included since fields are set idependently
+            fields = {'b': {'include': ...}}
+
+    class Child(Parent):
+        class Config:
+            # b is still included even if it doesn't occur here since fields
+            # are still considered separately.
+            # s however, is merged, resulting in only s1 being included.
+            fields = {'a': {'include': ...}, 's': {'include': {'s1'}}}
+
+    actual = Child(a=0, b=1, c=2).dict()
+    expected = {'a': 0, 'b': 1, 's': {'s1': 'v1'}}
+    assert actual == expected, 'Unexpected model export result'
 
 
 def test_custom_init_subclass_params():
@@ -1596,6 +1926,19 @@ def test_allow_mutation_field():
         r.id = 2
 
 
+def test_repr_field():
+    class Model(BaseModel):
+        a: int = Field()
+        b: int = Field(repr=True)
+        c: int = Field(repr=False)
+
+    m = Model(a=1, b=2, c=3)
+    assert repr(m) == 'Model(a=1, b=2)'
+    assert repr(m.__fields__['a'].field_info) == 'FieldInfo(default=PydanticUndefined, extra={})'
+    assert repr(m.__fields__['b'].field_info) == 'FieldInfo(default=PydanticUndefined, extra={})'
+    assert repr(m.__fields__['c'].field_info) == 'FieldInfo(default=PydanticUndefined, repr=False, extra={})'
+
+
 def test_inherited_model_field_copy():
     """It should copy models used as fields by default"""
 
@@ -1696,6 +2039,27 @@ def test_typing_coercion_dict():
     assert repr(m) == "Model(x={'one': 1, 'two': 2})"
 
 
+@pytest.mark.skipif(sys.version_info < (3, 7), reason='generic classes need 3.7')
+def test_typing_non_coercion_of_dict_subclasses():
+    KT = TypeVar('KT')
+    VT = TypeVar('VT')
+
+    class MyDict(Dict[KT, VT]):
+        def __repr__(self):
+            return f'MyDict({super().__repr__()})'
+
+    class Model(BaseModel):
+        a: MyDict
+        b: MyDict[str, int]
+        c: Dict[str, int]
+        d: Mapping[str, int]
+
+    assert (
+        repr(Model(a=MyDict({'a': 1}), b=MyDict({'a': '1'}), c=MyDict({'a': '1'}), d=MyDict({'a': '1'})))
+        == "Model(a=MyDict({'a': 1}), b=MyDict({'a': 1}), c={'a': 1}, d=MyDict({'a': 1}))"
+    )
+
+
 def test_typing_coercion_defaultdict():
     class Model(BaseModel):
         x: DefaultDict[int, str]
@@ -1707,6 +2071,30 @@ def test_typing_coercion_defaultdict():
     assert repr(m) == "Model(x=defaultdict(<class 'str'>, {1: '', 'a': ''}))"
 
 
+def test_typing_coercion_counter():
+    class Model(BaseModel):
+        x: Counter[str]
+
+    assert Model.__fields__['x'].type_ is int
+    assert repr(Model(x={'a': 10})) == "Model(x=Counter({'a': 10}))"
+
+
+def test_typing_counter_value_validation():
+    class Model(BaseModel):
+        x: Counter[str]
+
+    with pytest.raises(ValidationError) as exc_info:
+        Model(x={'a': 'a'})
+
+    assert exc_info.value.errors() == [
+        {
+            'loc': ('x', 'a'),
+            'msg': 'value is not a valid integer',
+            'type': 'type_error.integer',
+        }
+    ]
+
+
 def test_class_kwargs_config():
     class Base(BaseModel, extra='forbid', alias_generator=str.upper):
         a: int
@@ -1723,6 +2111,13 @@ def test_class_kwargs_config():
     assert Model.__fields__['b'].alias == 'B'  # alias_generator still works
 
 
+def test_class_kwargs_config_json_encoders():
+    class Model(BaseModel, json_encoders={int: str}):
+        pass
+
+    assert Model.__config__.json_encoders == {int: str}
+
+
 def test_class_kwargs_config_and_attr_conflict():
 
     with pytest.raises(
@@ -1734,3 +2129,32 @@ def test_class_kwargs_config_and_attr_conflict():
 
             class Config:
                 extra = 'forbid'
+
+
+def test_class_kwargs_custom_config():
+    class Base(BaseModel):
+        class Config(BaseConfig):
+            some_config = 'value'
+
+    class Model(Base, some_config='new_value'):
+        a: int
+
+    assert Model.__config__.some_config == 'new_value'
+
+
+@pytest.mark.skipif(sys.version_info < (3, 10), reason='need 3.10 version')
+def test_new_union_origin():
+    """On 3.10+, origin of `int | str` is `types.UnionType`, not `typing.Union`"""
+
+    class Model(BaseModel):
+        x: int | str
+
+    assert Model(x=3).x == 3
+    assert Model(x='3').x == 3
+    assert Model(x='pika').x == 'pika'
+    assert Model.schema() == {
+        'title': 'Model',
+        'type': 'object',
+        'properties': {'x': {'title': 'X', 'anyOf': [{'type': 'integer'}, {'type': 'string'}]}},
+        'required': ['x'],
+    }
