@@ -1,13 +1,21 @@
 import os
 import sys
 import uuid
+from datetime import datetime, timezone
 from pathlib import Path
-from typing import Dict, List, Optional, Set
+from typing import Any, Dict, List, Optional, Set, Tuple, Union
 
 import pytest
 
 from pydantic import BaseModel, BaseSettings, Field, HttpUrl, NoneStr, SecretStr, ValidationError, dataclasses
-from pydantic.env_settings import SettingsError, read_env_file
+from pydantic.env_settings import (
+    EnvSettingsSource,
+    InitSettingsSource,
+    SecretsSettingsSource,
+    SettingsError,
+    SettingsSourceCallable,
+    read_env_file,
+)
 
 try:
     import dotenv
@@ -71,7 +79,7 @@ def test_nested_env_with_basemodel(env):
     assert s.top == {'apple': 'value', 'banana': 'secret_value'}
 
 
-def test_nested_env_with_dict(env):
+def test_merge_dict(env):
     class Settings(BaseSettings):
         top: Dict[str, str]
 
@@ -82,6 +90,80 @@ def test_nested_env_with_dict(env):
     assert s.top == {'apple': 'value', 'banana': 'secret_value'}
 
 
+def test_nested_env_delimiter(env):
+    class SubSubValue(BaseSettings):
+        v6: str
+
+    class SubValue(BaseSettings):
+        v4: str
+        v5: int
+        sub_sub: SubSubValue
+
+    class TopValue(BaseSettings):
+        v1: str
+        v2: str
+        v3: str
+        sub: SubValue
+
+    class Cfg(BaseSettings):
+        v0: str
+        v0_union: Union[SubValue, int]
+        top: TopValue
+
+        class Config:
+            env_nested_delimiter = '__'
+
+    env.set('top', '{"v1": "json-1", "v2": "json-2", "sub": {"v5": "xx"}}')
+    env.set('top__sub__v5', '5')
+    env.set('v0', '0')
+    env.set('top__v2', '2')
+    env.set('top__v3', '3')
+    env.set('v0_union', '0')
+    env.set('top__sub__sub_sub__v6', '6')
+    env.set('top__sub__v4', '4')
+    cfg = Cfg()
+    assert cfg.dict() == {
+        'v0': '0',
+        'v0_union': 0,
+        'top': {
+            'v1': 'json-1',
+            'v2': '2',
+            'v3': '3',
+            'sub': {'v4': '4', 'v5': 5, 'sub_sub': {'v6': '6'}},
+        },
+    }
+
+
+def test_nested_env_delimiter_complex_required(env):
+    class Cfg(BaseSettings):
+        v: str = 'default'
+
+        class Config:
+            env_nested_delimiter = '__'
+
+    env.set('v__x', 'x')
+    env.set('v__y', 'y')
+    cfg = Cfg()
+    assert cfg.dict() == {'v': 'default'}
+
+
+def test_nested_env_delimiter_aliases(env):
+    class SubModel(BaseSettings):
+        v1: str
+        v2: str
+
+    class Cfg(BaseSettings):
+        sub_model: SubModel
+
+        class Config:
+            fields = {'sub_model': {'env': ['foo', 'bar']}}
+            env_nested_delimiter = '__'
+
+    env.set('foo__v1', '-1-')
+    env.set('bar__v2', '-2-')
+    assert Cfg().dict() == {'sub_model': {'v1': '-1-', 'v2': '-2-'}}
+
+
 class DateModel(BaseModel):
     pips: bool = False
 
@@ -387,22 +469,6 @@ def test_case_sensitive(monkeypatch):
     assert exc_info.value.errors() == [{'loc': ('foo',), 'msg': 'field required', 'type': 'value_error.missing'}]
 
 
-def test_case_insensitive(monkeypatch):
-    class Settings1(BaseSettings):
-        foo: str
-
-    with pytest.warns(DeprecationWarning, match='Settings2: "case_insensitive" is deprecated on BaseSettings'):
-
-        class Settings2(BaseSettings):
-            foo: str
-
-            class Config:
-                case_insensitive = False
-
-    assert Settings1.__config__.case_sensitive is False
-    assert Settings2.__config__.case_sensitive is True
-
-
 def test_nested_dataclass(env):
     @dataclasses.dataclass
     class MyDataclass:
@@ -424,8 +490,15 @@ def test_env_takes_precedence(env):
         foo: int
         bar: str
 
-        def _build_values(self, init_kwargs, _env_file, _env_file_encoding, _secrets_dir):
-            return {**init_kwargs, **self._build_environ()}
+        class Config:
+            @classmethod
+            def customise_sources(
+                cls,
+                init_settings: SettingsSourceCallable,
+                env_settings: SettingsSourceCallable,
+                file_secret_settings: SettingsSourceCallable,
+            ) -> Tuple[SettingsSourceCallable, ...]:
+                return env_settings, init_settings
 
     env.set('BAR', 'env setting')
 
@@ -439,24 +512,71 @@ def test_config_file_settings_nornir(env):
     See https://github.com/samuelcolvin/pydantic/pull/341#issuecomment-450378771
     """
 
+    def nornir_settings_source(settings: BaseSettings) -> Dict[str, Any]:
+        return {'param_a': 'config a', 'param_b': 'config b', 'param_c': 'config c'}
+
     class Settings(BaseSettings):
         param_a: str
         param_b: str
         param_c: str
 
-        def _build_values(self, init_kwargs, _env_file, _env_file_encoding, _secrets_dir):
-            config_settings = init_kwargs.pop('__config_settings__')
-            return {**config_settings, **init_kwargs, **self._build_environ()}
+        class Config:
+            @classmethod
+            def customise_sources(
+                cls,
+                init_settings: SettingsSourceCallable,
+                env_settings: SettingsSourceCallable,
+                file_secret_settings: SettingsSourceCallable,
+            ) -> Tuple[SettingsSourceCallable, ...]:
+                return env_settings, init_settings, nornir_settings_source
 
     env.set('PARAM_C', 'env setting c')
 
-    config = {'param_a': 'config a', 'param_b': 'config b', 'param_c': 'config c'}
-    s = Settings(__config_settings__=config, param_b='argument b', param_c='argument c')
+    s = Settings(param_b='argument b', param_c='argument c')
     assert s.param_a == 'config a'
     assert s.param_b == 'argument b'
     assert s.param_c == 'env setting c'
 
 
+def test_env_union_with_complex_subfields_parses_json(env):
+    class A(BaseSettings):
+        a: str
+
+    class B(BaseSettings):
+        b: int
+
+    class Settings(BaseSettings):
+        content: Union[A, B, int]
+
+    env.set('content', '{"a": "test"}')
+    s = Settings()
+    assert s.content == A(a='test')
+
+
+def test_env_union_with_complex_subfields_parses_plain_if_json_fails(env):
+    class A(BaseSettings):
+        a: str
+
+    class B(BaseSettings):
+        b: int
+
+    class Settings(BaseSettings):
+        content: Union[A, B, datetime]
+
+    env.set('content', '2020-07-05T00:00:00Z')
+    s = Settings()
+    assert s.content == datetime(2020, 7, 5, 0, 0, tzinfo=timezone.utc)
+
+
+def test_env_union_without_complex_subfields_does_not_parse_json(env):
+    class Settings(BaseSettings):
+        content: Union[datetime, str]
+
+    env.set('content', '2020-07-05T00:00:00Z')
+    s = Settings()
+    assert s.content == datetime(2020, 7, 5, 0, 0, tzinfo=timezone.utc)
+
+
 test_env_file = """\
 # this is a comment
 A=good string
@@ -784,6 +904,33 @@ def test_secrets_path_url(tmp_path):
     assert Settings().dict() == {'foo': 'http://www.example.com', 'bar': SecretStr('snap')}
 
 
+def test_secrets_path_json(tmp_path):
+    p = tmp_path / 'foo'
+    p.write_text('{"a": "b"}')
+
+    class Settings(BaseSettings):
+        foo: Dict[str, str]
+
+        class Config:
+            secrets_dir = tmp_path
+
+    assert Settings().dict() == {'foo': {'a': 'b'}}
+
+
+def test_secrets_path_invalid_json(tmp_path):
+    p = tmp_path / 'foo'
+    p.write_text('{"a": "b"')
+
+    class Settings(BaseSettings):
+        foo: Dict[str, str]
+
+        class Config:
+            secrets_dir = tmp_path
+
+    with pytest.raises(SettingsError, match='error parsing JSON for "foo"'):
+        Settings()
+
+
 def test_secrets_missing(tmp_path):
     class Settings(BaseSettings):
         foo: str
@@ -813,12 +960,10 @@ def test_secrets_invalid_secrets_dir(tmp_path):
 @pytest.mark.skipif(sys.platform.startswith('win'), reason='windows paths break regex')
 def test_secrets_missing_location(tmp_path):
     class Settings(BaseSettings):
-        foo: str
-
         class Config:
             secrets_dir = tmp_path / 'does_not_exist'
 
-    with pytest.raises(SettingsError, match=f'directory "{tmp_path}/does_not_exist" does not exist'):
+    with pytest.warns(UserWarning, match=f'directory "{tmp_path}/does_not_exist" does not exist'):
         Settings()
 
 
@@ -852,3 +997,93 @@ def test_secrets_dotenv_precedence(tmp_path):
             secrets_dir = tmp_path
 
     assert Settings(_env_file=e).dict() == {'foo': 'foo_env_value_str'}
+
+
+def test_external_settings_sources_precedence(env):
+    def external_source_0(settings: BaseSettings) -> Dict[str, str]:
+        return {'apple': 'value 0', 'banana': 'value 2'}
+
+    def external_source_1(settings: BaseSettings) -> Dict[str, str]:
+        return {'apple': 'value 1', 'raspberry': 'value 3'}
+
+    class Settings(BaseSettings):
+        apple: str
+        banana: str
+        raspberry: str
+
+        class Config:
+            @classmethod
+            def customise_sources(
+                cls,
+                init_settings: SettingsSourceCallable,
+                env_settings: SettingsSourceCallable,
+                file_secret_settings: SettingsSourceCallable,
+            ) -> Tuple[SettingsSourceCallable, ...]:
+                return init_settings, env_settings, file_secret_settings, external_source_0, external_source_1
+
+    env.set('banana', 'value 1')
+    assert Settings().dict() == {'apple': 'value 0', 'banana': 'value 1', 'raspberry': 'value 3'}
+
+
+def test_external_settings_sources_filter_env_vars():
+    vault_storage = {'user:password': {'apple': 'value 0', 'banana': 'value 2'}}
+
+    class VaultSettingsSource:
+        def __init__(self, user: str, password: str):
+            self.user = user
+            self.password = password
+
+        def __call__(self, settings: BaseSettings) -> Dict[str, str]:
+            vault_vars = vault_storage[f'{self.user}:{self.password}']
+            return {
+                field.alias: vault_vars[field.name]
+                for field in settings.__fields__.values()
+                if field.name in vault_vars
+            }
+
+    class Settings(BaseSettings):
+        apple: str
+        banana: str
+
+        class Config:
+            @classmethod
+            def customise_sources(
+                cls,
+                init_settings: SettingsSourceCallable,
+                env_settings: SettingsSourceCallable,
+                file_secret_settings: SettingsSourceCallable,
+            ) -> Tuple[SettingsSourceCallable, ...]:
+                return (
+                    init_settings,
+                    env_settings,
+                    file_secret_settings,
+                    VaultSettingsSource(user='user', password='password'),
+                )
+
+    assert Settings().dict() == {'apple': 'value 0', 'banana': 'value 2'}
+
+
+def test_customise_sources_empty():
+    class Settings(BaseSettings):
+        apple: str = 'default'
+        banana: str = 'default'
+
+        class Config:
+            @classmethod
+            def customise_sources(cls, *args, **kwargs):
+                return ()
+
+    assert Settings().dict() == {'apple': 'default', 'banana': 'default'}
+    assert Settings(apple='xxx').dict() == {'apple': 'default', 'banana': 'default'}
+
+
+def test_builtins_settings_source_repr():
+    assert (
+        repr(InitSettingsSource(init_kwargs={'apple': 'value 0', 'banana': 'value 1'}))
+        == "InitSettingsSource(init_kwargs={'apple': 'value 0', 'banana': 'value 1'})"
+    )
+    assert (
+        repr(EnvSettingsSource(env_file='.env', env_file_encoding='utf-8'))
+        == "EnvSettingsSource(env_file='.env', env_file_encoding='utf-8', env_nested_delimiter=None)"
+    )
+    assert repr(SecretsSettingsSource(secrets_dir='/secrets')) == "SecretsSettingsSource(secrets_dir='/secrets')"
